@@ -11,6 +11,19 @@ from InSpice.Spice.Netlist import Circuit, SubCircuit
 from InSpice.Spice.Parser.HighLevelParser import SpiceSource
 from InSpice.Spice.Parser.Translator import Builder
 
+# package download dependencies
+import urllib.request
+import shutil
+import tempfile
+import hashlib
+from pathlib import Path
+from urllib.parse import urlparse
+try:
+    import py7zr
+    shutil.register_unpack_format('7zip', ['.7z'], py7zr.unpack_7zarchive)
+except ImportError:
+    pass
+
 
 def model_key(bare_id):
     """
@@ -226,8 +239,26 @@ def netlist(docs, models):
 class NyanCADMixin:
     """Mixin providing NyanCAD integration for InSpice netlist objects."""
     
+    def _select_template(self, templates, sim):
+        """Select template from templates dict based on sim parameter.
+        
+        Returns:
+            dict: selected template or None if no templates
+        """
+        spice_templates = templates.get('spice', [])
+        if not spice_templates:
+            return None
+            
+        # Look for implementation matching sim parameter, fallback to first (default)
+        for template in spice_templates:
+            if template.get('name', '').lower() == sim.lower():
+                return template
+        
+        return spice_templates[0]  # Use first as default
+    
     def populate_from_nyancad(self, docs, models, corner='tt', sim='NgSpice'):
         """Populate this netlist with elements from NyanCAD docs."""
+        self.used_models = set()
         nl = netlist(docs, models)
         
         for dev_id, ports in nl.items():
@@ -241,49 +272,92 @@ class NyanCADMixin:
         props = dev.get('props', {}).copy()
 
         model_id = model_key(dev.get('model'))
+        model_use_x = False
+        model_name = None
+
         if model_id and model_id in models:
-            props['model'] = models[model_id]['name']
-        
+            self.used_models.add(model_id)
+            model_def = models[model_id]
+            model_name = model_def['name']
+            props['model'] = model_name
+            
+            # Check if model should be used as subcircuit (X) instead of component
+            templates = model_def.get('templates', {})
+            selected_template = self._select_template(templates, sim)
+            model_use_x = selected_template.get('use-x', False) if selected_template else False
+
+        print(props)
         # Helper to get port by name
         def p(port_name):
             return ports[port_name]
         
-        
         # Map device types to InSpice API methods
         if device_type == "resistor":
-            resistance = props.get('resistance')
-            self.R(name, p('P'), p('N'), resistance)
+            if model_use_x:
+                subcircuit_model = props.pop('model', model_name)
+                self.X(name, subcircuit_model, p('P'), p('N'), **props)
+            else:
+                resistance = props.get('resistance')
+                self.R(name, p('P'), p('N'), resistance)
             
         elif device_type == "capacitor":
-            capacitance = props.get('capacitance')
-            self.C(name, p('P'), p('N'), capacitance)
+            if model_use_x:
+                subcircuit_model = props.pop('model', model_name)
+                self.X(name, subcircuit_model, p('P'), p('N'), **props)
+            else:
+                capacitance = props.get('capacitance')
+                self.C(name, p('P'), p('N'), capacitance)
             
         elif device_type == "inductor":
-            inductance = props.get('inductance')
-            self.L(name, p('P'), p('N'), inductance)
+            if model_use_x:
+                subcircuit_model = props.pop('model', model_name)
+                self.X(name, subcircuit_model, p('P'), p('N'), **props)
+            else:
+                inductance = props.get('inductance')
+                self.L(name, p('P'), p('N'), inductance)
             
         elif device_type == "diode":
-            self.D(name, p('P'), p('N'), **props)
+            if model_use_x:
+                subcircuit_model = props.pop('model', model_name)
+                self.X(name, subcircuit_model, p('P'), p('N'), **props)
+            else:
+                self.D(name, p('P'), p('N'), **props)
             
         elif device_type == "vsource":
-            dc = props.get('dc')
-            ac = props.get('ac')
-            tran = props.get('tran')
-            self.V(name, p('P'), p('N'), dc, ac, tran)
+            if model_use_x:
+                subcircuit_model = props.pop('model', model_name)
+                self.X(name, subcircuit_model, p('P'), p('N'), **props)
+            else:
+                dc = props.get('dc')
+                ac = props.get('ac')
+                tran = props.get('tran')
+                self.V(name, p('P'), p('N'), dc, ac, tran)
             
         elif device_type == "isource":
-            dc = props.get('dc')
-            ac = props.get('ac')
-            tran = props.get('tran')
-            self.I(name, p('P'), p('N'), dc, ac, tran)
+            if model_use_x:
+                subcircuit_model = props.pop('model', model_name)
+                self.X(name, subcircuit_model, p('P'), p('N'), **props)
+            else:
+                dc = props.get('dc')
+                ac = props.get('ac')
+                tran = props.get('tran')
+                self.I(name, p('P'), p('N'), dc, ac, tran)
             
         elif device_type in {"pmos", "nmos"}:
-            bulk_node = p('B') if 'B' in ports else self.gnd
-            # Map parameter names and copy dict to avoid mutation
-            self.M(name, p('D'), p('G'), p('S'), bulk_node, **props)
+            if model_use_x:
+                subcircuit_model = props.pop('model', model_name)
+                bulk_node = p('B') if 'B' in ports else self.gnd
+                self.X(name, subcircuit_model, p('D'), p('G'), p('S'), bulk_node, **props)
+            else:
+                bulk_node = p('B') if 'B' in ports else self.gnd
+                self.M(name, p('D'), p('G'), p('S'), bulk_node, **props)
             
         elif device_type in {"npn", "pnp"}:
-            self.Q(name, p('C'), p('B'), p('E'), **props)
+            if model_use_x:
+                subcircuit_model = props.pop('model', model_name)
+                self.X(name, subcircuit_model, p('C'), p('B'), p('E'), **props)
+            else:
+                self.Q(name, p('C'), p('B'), p('E'), **props)
             
         else:  # subcircuit
             if model_id in models:
@@ -312,8 +386,12 @@ class NyanCircuit(NyanCADMixin, Circuit):
         
         models = schem["models"]
         
-        # Process models: create subcircuits for schematic models, add SPICE for others
-        for model_key_str, model_def in models.items():
+        # First populate main circuit elements to collect used models
+        self.populate_from_nyancad(schem[name], models, corner, sim)
+        
+        # Then process only the used models: create subcircuits for schematic models, add SPICE for others
+        for model_key_str in self.used_models:
+            model_def = models[model_key_str]
             # Extract bare model ID for schematic lookup (models dict keys always have "models:" prefix)
             model_id = bare_id(model_key_str)
             # Skip the top-level circuit itself
@@ -328,23 +406,11 @@ class NyanCircuit(NyanCADMixin, Circuit):
                 else:
                     # Add SPICE code for models with templates (SPICE subcircuits)
                     templates = model_def.get('templates', {})
-                    spice_templates = templates.get('spice', [])
-                    if spice_templates:
-                        # Look for implementation matching sim parameter, fallback to first (default)
-                        selected_template = None
-                        for template in spice_templates:
-                            if template.get('name', '').lower() == sim.lower():
-                                selected_template = template
-                                break
-                        if not selected_template:
-                            selected_template = spice_templates[0]  # Use first as default
-                        
+                    selected_template = self._select_template(templates, sim)
+                    if selected_template:
                         code = selected_template.get('code', '').strip()
                         if code:
                             self.add_spice_code(code)
-        
-        # Populate main circuit elements
-        self.populate_from_nyancad(schem[name], models, corner, sim)
     
     def add_spice_code(self, spice_code: str):
         """Add SPICE code to circuit. Try structured parsing first, fallback to raw injection.
@@ -355,15 +421,85 @@ class NyanCircuit(NyanCADMixin, Circuit):
         try:
             # Parse SPICE code
             spice_source = SpiceSource(spice_code, title_line=False)
+            
+            # Resolve URL includes in the SpiceSource before building
+            self.resolve_url_includes(spice_source)
+            
             builder = Builder()
             parsed_circuit = builder.translate(spice_source)
             # Copy all content to self (models, subcircuits, elements)
             parsed_circuit.copy_to(self)
+            # copy includes and parameters
+            for include in parsed_circuit._includes:
+                self.include(include)
+            for path, section in parsed_circuit._libs:
+                self.lib(path, section)
+            for name, value in parsed_circuit._parameters.items():
+                self.parameter(name, value)
+
         except Exception as e:
             print(f"SPICE parsing failed: {e}")
             print("Falling back to raw SPICE injection")
             # Append to raw_spice
             self.raw_spice += '\n' + spice_code.strip() + '\n'
+    
+    def _resolve_url_path(self, path_obj):
+        """Helper to resolve a single URL path to a local file path."""
+        cache_dir = Path(tempfile.gettempdir()) / "nyancad_archive_cache"
+        cache_dir.mkdir(exist_ok=True)
+        
+        # Get the raw path from the AST
+        ast = path_obj._ast
+        path_str = str(next(iter(ast))).strip("\"'")
+
+        parsed = urlparse(path_str)
+        
+        if parsed.scheme in ('http', 'https'):
+            try:
+                archive_url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
+                entrypoint = parsed.fragment
+                
+                # Generate cache key from URL
+                url_hash = hashlib.md5(archive_url.encode()).hexdigest()[:8]
+                filename = Path(parsed.path).name
+                cached_file = cache_dir / f"{url_hash}_{filename}"
+                
+                # Download if not cached
+                if not cached_file.exists():
+                    print(f"Downloading: {archive_url}")
+                    urllib.request.urlretrieve(archive_url, cached_file)
+                
+                if entrypoint:
+                    # Archive with entrypoint - extract and resolve
+                    base_name = cached_file.stem  # Get filename without extension
+                    extract_dir = cache_dir / base_name
+                    if not extract_dir.exists():
+                        shutil.unpack_archive(str(cached_file), str(extract_dir))
+                    
+                    resolved_path = extract_dir / entrypoint
+                    if not resolved_path.exists():
+                        print(f"Warning: Entrypoint not found in archive: {entrypoint}")
+                        return
+                else:
+                    # Bare SPICE file - use directly
+                    resolved_path = cached_file
+                
+                # Replace the _path in the object
+                path_obj._path = resolved_path
+            except Exception as e:
+                print(f"Failed to resolve path {path_str}: {e}")
+                # Leave original path unchanged on error
+
+    def resolve_url_includes(self, spice_source):
+        """Resolve HTTP includes (archives or bare files) to local file paths in a SpiceSource object."""
+        # Process includes
+        for include in spice_source._includes:
+            self._resolve_url_path(include)
+        
+        # Also process libs if they exist
+        if hasattr(spice_source, '_libs'):
+            for lib in spice_source._libs:
+                self._resolve_url_path(lib)
 
 
 class NyanSubCircuit(NyanCADMixin, SubCircuit):
